/*
 * File Archive Library
 * Copyright (C) 2000-2001 Patrizio Bruno <patrizio@dada.it>
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    Read the file LICENSE for more.
 *
 */

/*
 * $Log: archive.c,v $
 * Revision 1.19  2001/05/25 15:28:35  rusty
 *
 * fixed the struct dirent handling on solaris
 *
 * Revision 1.18  2001/05/24 16:59:33  rusty
 *
 * alphasort's wrapper
 *
 * Revision 1.17  2001/05/24 16:48:16  rusty
 *
 * scandir wrapper for systems that do not natively support this BSD function
 *
 * Revision 1.16  2001/05/21 12:16:45  rusty
 *
 * added a md5 checksum for the whole archive
 *
 * Revision 1.15  2001/05/21 11:06:31  rusty
 *
 * byte ordering fixed
 *
 * Revision 1.14  2001/05/16 16:08:17  rusty
 * *** empty log message ***
 *
 * Revision 1.13  2001/05/16 11:06:58  rusty
 * *** empty log message ***
 *
 * Revision 1.12  2001/05/16 11:04:57  rusty
 *
 * some debug code removed
 * added a function displaying the internal version of the library: far_version() in utils.c
 *
 * Revision 1.11  2001/05/16 10:50:39  rusty
 *
 * fixed the checksum control on archive expand
 * some little code clean up
 *
 * Revision 1.10  2000/09/21 15:25:24  rusty
 *
 * some verboseness added
 *
 * Revision 1.9  2000/09/21 11:27:44  rusty
 *
 * bug fixes
 *
 * Revision 1.8  2000/09/20 17:43:07  rusty
 *
 * why checksums of big files seem to be corrupted?
 *
 * Revision 1.7  2000/09/20 16:59:06  rusty
 * *** empty log message ***
 *
 * Revision 1.6  2000/09/20 16:07:38  rusty
 *
 * some more free() stuff
 *
 * Revision 1.5  2000/09/20 14:59:00  rusty
 *
 * free_info_list()'s "Segmentation fault" fixed
 *
 * Revision 1.4  2000/09/20 14:10:22  rusty
 *
 * First working (basic) release
 *
 * Revision 1.3  2000/09/20 11:48:00  rusty
 *
 * lot of bug fixes, now the library seemes to archive correctly
 *
 * TODO:
 *   - archive extract
 *   - offsets in the FAR_INFO structure
 *   - checksums for files and for the whole archive
 *
 * Revision 1.2  2000/09/20 08:30:07  rusty
 * *** empty log message ***
 *
 * Revision 1.1.1.1  2000/09/19 18:34:08  rusty
 * First release of the File ARchiver
 *
 */

#include "far.h"
#include "archive.h"
#include "utils.h"

char *arch_version = "$Revision: 1.19 $";

void *far_create(char *path, char *reg_exp, size_t *bufsize)
{
	FAR_INFO		*info	= (FAR_INFO *)NULL;
	unsigned	char	*result	= (unsigned char *)NULL,
				*buffer	= (unsigned char *)NULL;
	char			*start	= (char *)NULL,
				**list	= (char **)NULL;
	int			n_files	= 0;

	if(path)
	{
		n_files = strcount((char *)path, " ");
		n_files++;
		list = calloc(1, ((n_files + 5) * sizeof(char *)));
		make_list((char *)path, " ", list, n_files);
	}
	else
	{
		list = calloc(2, 1);
		list[0] = strdup("");
	}

	start = getcwd(NULL, 1024);
	(*(bufsize)) = 0;

	info = far_scanpath(start, list, reg_exp, TRUE);

	if(reg_exp)
	{
		regfree(&preg);
	}

	buffer = far_build_archive(info, bufsize);
	result = far_end_create(info, buffer, bufsize);

	free(buffer);
	far_free_infolist(info);

	return (void *)result;
}

FAR_INFO *far_scanpath(char *start, char **path_src,
			char *reg_exp,  int first)
{
	FAR_INFO	*result		= (FAR_INFO *)NULL,
			*tmp_res	= (FAR_INFO *)NULL,
			*prev		= (FAR_INFO *)NULL,
			*ptr		= (FAR_INFO *)NULL;
	struct	dirent	**namelist	= (struct dirent **)NULL;
	int		count		= 0,
			n_entries	= 0,
			init		= 0;
	static	int	(*select)(const struct dirent *);
	char		*path		= (char *)NULL;

	if(reg_exp)
	{
		far_regexp_set(reg_exp);
		select = far_regexp_select;
	}
	else
	{
		select = NULL;
	}

	for(; (*path_src) ; path_src++)
	{
		if(*(*path_src))
		{
			if((*(*path_src)) &&
				(((*path_src)[strlen(*path_src)-1]) != '/') &&
				is_adir(*path_src))
			{
				path = calloc(1, strlen(*path_src) + 2);
				strcpy(path, (*path_src));
				strcat(path, "/");
			}
			else
			{
				path = strdup(*path_src);
			}
		}

		if(!(path))
		{
			first = FALSE;
		}
		else if(!(*path))
		{
			first = FALSE;
		}
		else if(!(strcmp(path, "./")))
		{
			first = FALSE;
		}

		path = ((path) ? path : strdup("./"));

		if(first)
		{
			tmp_res = far_getpath_info(start, path, reg_exp);

			if(tmp_res && !(result))
			{
				ptr = result = tmp_res;
				ptr = far_find_last(ptr);
			}
			else if(tmp_res && result)
			{
				ptr->next = tmp_res;
				ptr->next->prev = ptr;
				ptr = far_find_last(ptr);
			}
			else
			{
				result = NULL;
			}
		}
		else
		{
			chdir(path);
			n_entries = far_scandir(".", &namelist, select,
						far_alphasort);
			if(n_entries != -1)
			{
				for(count = 0 ; count < n_entries ; count++)
				{
					if(!(strcmp(namelist[count]->d_name,
							".")))
					{
						init = count + 1;
						continue;
					}

					if(!(strcmp(namelist[count]->d_name,
							"..")))
					{
						init = count + 1;
						continue;
					}

					tmp_res = far_getpath_info(start,
						namelist[count]->d_name,
						reg_exp);

					if(tmp_res)
					{
						if((count == init) && !(result))
						{
							ptr = result = tmp_res;
							ptr = far_find_last(ptr);
						}
						else
						{
							ptr->next = tmp_res;
							prev = ptr;
							ptr = ptr->next;
							ptr->prev = prev;
							ptr = far_find_last(ptr);
						}
					}
					else
					{
						result = NULL;
					}
				}
			}
		}

		if(path)
		{
			free(path);
		}
	}

	return result;
}

FAR_INFO *far_getpath_info(char *start, char *path,
			char *reg_exp)
{
	FAR_INFO	*result		= (FAR_INFO *)NULL;
	struct	stat	stat_buf;
	int		check		= 0;
	char		*cwd		= (char *)NULL;

	if((check = lstat(path, &stat_buf) != -1))
	{
		if(!(S_ISDIR(stat_buf.st_mode)))
		{
			result = far_get_info(start, path);
		}
		else
		{
			result = far_get_info(start, path);

			if(result)
			{
				char	*tpath[2]	= { 0 };

				tpath[0] = (char *)path;
				cwd = getcwd(NULL, 1024);
				result->next = far_scanpath(start, tpath,
							reg_exp, FALSE);
				chdir(cwd);
			}
		}
	}
	else
	{
		result = NULL;
	}

	return result;
}

FAR_INFO *far_get_info(char *start, char *path)
{
	FAR_INFO	*result	= (FAR_INFO *)NULL;
	char		*pwd	= (char *)NULL;

	result = calloc(1, sizeof(FAR_INFO));

	if(lstat(path, &(result->entry_info)) != -1)
	{
		pwd = getcwd(NULL, 1024);

		if(strcmp(pwd, start) && (path[0] != '/'))
		{
			sprintf(result->path, (pwd + (strlen(start) + 1)));
			strcat(result->path, "/");
		}

		strncat(result->path, path, (sizeof(result->path) - 1));
	}
	else
	{
		free(result);
		result = NULL;
	}

	return result;
}

void *far_build_archive(FAR_INFO *info, size_t *bufsize)
{
	unsigned	char	*result	= (unsigned char *)NULL,
				*last	= (unsigned char *)NULL;
	FAR_INFO		*ptr	= info;

	(*bufsize) = far_get_total_size(info);
	last = result = calloc(1, (*bufsize));

	while(ptr)
	{
		if(S_ISREG(ptr->entry_info.st_mode))
		{
			last = far_append_file(ptr, last);
		}
		else if(S_ISLNK(ptr->entry_info.st_mode))
		{
			if(far_set_link(ptr) == -1)
			{
				last = NULL;
			}
		}

		if(!(last))
		{
			free(result);
			result = NULL;
			break;
		}

		far_printf(stderr, "%s\n", ptr->path);
		ptr = ptr->next;
	}

	return (void *)result;
}

void *far_append_file(FAR_INFO *info, void *current)
{
	unsigned	char	*result	= (unsigned char *)NULL,
				*curr	= (unsigned char *)current;
	int			fd		= 0;
	md5_state_t		md5;

	fd = open(info->path, O_RDONLY);

	if(fd != -1)
	{
		if(read(fd, curr, info->entry_info.st_size) != -1)
		{
			result = curr + info->entry_info.st_size;
		}

		close(fd);

		md5_init(&md5);
		md5_append(&md5, curr, info->entry_info.st_size);
		md5_finish(&md5, info->chksum);
	}

	return result;
}

int far_set_link(FAR_INFO *link_info)
{
	return readlink(link_info->path, link_info->link_path,
			(sizeof(link_info->link_path) - 1));
}

size_t far_get_total_size(FAR_INFO *info)
{
	FAR_INFO	*ptr	= (FAR_INFO *)info;
	size_t		result	= 0,
			offset	= 0;

	while(ptr)
	{
		if(S_ISREG(ptr->entry_info.st_mode))
		{
			result += ptr->entry_info.st_size;
			ptr->offset = offset;
			offset += ptr->entry_info.st_size;
		}

		ptr = ptr->next;
	}

	return result;
}

void far_free_infolist(FAR_INFO *list)
{
	FAR_INFO	*ptr	= (FAR_INFO *)list,
			*ptr2	= (FAR_INFO *)list;

	while(ptr2)
	{
		ptr = ptr->next;
		far_free_info(ptr2);
		ptr2 = ptr;
	}

	return;
}

void far_free_info(FAR_INFO *info)
{
	if(info)
	{
		free(info);
	}

	return;
}

int far_regexp_select(const struct dirent *dir)
{
	int		result	= REG_NOMATCH;
	regmatch_t	pmatch[1];

	result = regexec(&preg, dir->d_name, 0, pmatch, 0);

	if(result != 0)
	{	
		result = 0;
	}
	else
	{
		result = 1;
	}

	return result;
}

FAR_INFO *far_find_last(FAR_INFO *list)
{
	FAR_INFO	*result	= (FAR_INFO *)list;

	while(result && result->next)
	{
		result = result->next;
	}

	return result;
}

FAR_INFO *far_find_first(FAR_INFO *list)
{
	FAR_INFO	*result	= (FAR_INFO *)list;

	while(result && result->prev)
	{
		result = result->prev;
	}

	return result;
}

int far_info_count(FAR_INFO *list)
{
	FAR_INFO	*ptr	= (FAR_INFO *)list;
	int		result	= 0;

	while(ptr)
	{
		result++;
		ptr = ptr->next;
	}

	return result;
}

void *far_end_create(FAR_INFO *info, void *buffer, size_t *bufsize)
{
	unsigned	char	*result		= (unsigned char *)NULL;
	size_t			totsize		= 0,
				headsize	= 0,
				count		= 0;
	FAR_INFO		*ptr		= (FAR_INFO *)info;
	unsigned	char	cksum[16]	= { 0 };
	md5_state_t		md5;

	headsize = (far_info_count(info) * sizeof(FAR_INFO));
	totsize = (headsize + sizeof(size_t) + (*bufsize) + 16);
	result = calloc(1, totsize);

	headsize = htonl(headsize);
	memcpy(result, (&(headsize)), sizeof(size_t));
	count = sizeof(size_t);

	while(ptr)
	{
		far_byteorder_conv(ptr);
		memcpy((result + count), ptr, sizeof(FAR_INFO));
		ptr = ptr->next;
		count += sizeof(FAR_INFO);
	}

	memcpy((result + count), buffer, (*bufsize));

	md5_init(&md5);
	md5_append(&md5, result, (totsize - 16));
	md5_finish(&md5, cksum);
	memcpy((result + (totsize - 16)), cksum, 16);

	(*bufsize) = totsize;

	return result;
}

